import warnings

import numpy as np

from stimupy.components.gaussians import gaussian
from stimupy.components.shapes import parallelogram, rectangle
from stimupy.stimuli.waves import sine_linear as sinewave
from stimupy.stimuli.waves import square_linear as squarewave
from stimupy.utils import pad_dict_to_shape, pad_dict_to_visual_size, resolution, strip_dict
from stimupy.utils.filters import convolve

__all__ = [
    "sinewave",
    "squarewave",
    "on_uniform",
    "on_grating",
    "on_grating_masked",
    "phase_shifted",
    "grating_induction",
    "grating_induction_blur",
]


def on_uniform(
    visual_size=None,
    ppd=None,
    shape=None,
    grating_size=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensity_bars=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    intensity_background=0.5,
    origin="corner",
    round_phase_width=True,
):
    """Spatial square-wave grating (set of bars), on a uniform background

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    grating_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_bars : Sequence[float, ...]
        intensity value for each bar, by default (1.0, 0.0).
        Can specify as many intensities as n_bars;
        If fewer intensities are passed than n_bars, cycles through intensities
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    intensity_background = float
        intensity value of background
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of bars given resolution

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)

    # Spatial square-wave grating
    stim = squarewave(
        visual_size=grating_size,
        ppd=ppd,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        intensity_bars=intensity_bars,
        target_indices=target_indices,
        intensity_target=intensity_target,
        origin=origin,
        round_phase_width=round_phase_width,
    )

    # Padding
    stim = pad_dict_to_visual_size(
        stim, visual_size=visual_size, ppd=ppd, pad_value=intensity_background
    )

    # Repack
    stim.update(
        intensity_background=intensity_background,
        grating_size=stim["visual_size"],
        grating_shape=stim["shape"],
        shape=stim["img"].shape,
        visual_size=visual_size,
        target_indices=target_indices,
        intensity_target=intensity_target,
    )

    return stim


def on_grating_masked(
    small_grating_params,
    large_grating_params,
    mask_size=None,
    mask_rotation=None,
):
    """Small grating, with a parallelogram-like shape, on a larger grating

    Parameters
    ----------
    small_grating_params : dict
        kwargs to generate small grating
    large_grating_params : dict
        kwargs to generate larger grating
    mask_size : Sequence[Number, Number, Number], Sequence[Number, Number], Number or None (default)
        size (height, width, depth) of parallelogram-like mask in degrees visual angle
    mask_rotation: float
        rotation of the mask in degree

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each bar (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """

    # Create gratings
    small_grating = squarewave(**small_grating_params)
    large_grating = squarewave(**large_grating_params)

    if small_grating["ppd"] != large_grating["ppd"]:
        raise ValueError("Gratings must have same ppd")

    if mask_size is None:
        mask_size = small_grating["visual_size"]
    if mask_rotation is None:
        mask_rotation = 0

    window = parallelogram(
        ppd=small_grating["ppd"],
        shape=large_grating["shape"],
        parallelogram_size=mask_size,
        rotation=mask_rotation,
    )["img"]

    small_grating = pad_dict_to_shape(small_grating, large_grating["shape"])
    window = window * (small_grating["pad_mask"] - 1)
    img = np.where(window, small_grating["img"], large_grating["img"])
    mask = np.where(window, small_grating["target_mask"], 0)

    stim = {
        "img": img,
        "visual_size": large_grating["visual_size"],
        "ppd": large_grating["ppd"],
        "target_mask": mask.astype(int),
        "small_grating_mask": small_grating["grating_mask"],
        "large_grating_mask": large_grating["grating_mask"],
        "small_grating_params": strip_dict(small_grating, squarewave),
        "large_grating_params": strip_dict(large_grating, squarewave),
        "mask_size": mask_size,
        "mask_rotation": mask_rotation,
    }
    return stim


def on_grating(
    small_grating_params,
    large_grating_params,
):
    """Small grating on a larger grating

    Parameters
    ----------
    small_grating_params : dict
        kwargs to generate small grating
    large_grating_params : dict
        kwargs to generate larger grating

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """

    stim = on_grating_masked(
        small_grating_params,
        large_grating_params,
    )
    return stim


def phase_shifted(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    distance_metric="horizontal",
    phase_shift=0,
    intensity_bars=(0.0, 1.0),
    target_size=None,
    target_phase_shift=0,
    intensity_target=0.5,
    origin="corner",
):
    if target_size is None:
        raise ValueError(
            "counterphase_induction() missing argument 'target_size' which is not 'None'"
        )
    if distance_metric == "horizontal":
        rotation = 0
    elif distance_metric == "vertical":
        rotation = 90
    else:
        raise ValueError("distance_metric must be horizontal or vertical")

    # Spatial square-wave grating
    stim = squarewave(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        rotation=rotation,
        phase_shift=phase_shift,
        period=period,
        intensity_bars=intensity_bars,
        origin=origin,
        round_phase_width=True,
    )

    stim_target = squarewave(
        visual_size=target_size,
        ppd=stim["ppd"],
        bar_width=stim["bar_width"],
        rotation=rotation,
        phase_shift=0,
        period=period,
        intensity_bars=(intensity_target, 0),
        origin=origin,
        round_phase_width=True,
    )
    stim_target = pad_dict_to_shape(stim_target, stim["shape"], 0)
    cycle_px = stim_target["bar_width"] * stim_target["ppd"][0] * 2

    # Translate phase information into pixels
    target_phasea = np.abs(target_phase_shift)
    target_phasea = target_phasea % 360
    target_amount = target_phasea / 360.0
    target_shift = target_amount * cycle_px
    target_shifti = int(np.round(target_shift))
    target_phasei = target_shifti / cycle_px * 360

    if target_shift != int(target_shift):
        s = np.sign(target_phase_shift)
        warnings.warn(f"Rounding phase; {target_phase_shift} -> {s*target_phasei}")

    # Shift targets by specified phase
    cy, cx = stim["shape"]
    if target_phase_shift < 0:
        if distance_metric == "horizontal":
            stim_target["img"][:, 0 : cx - target_shifti] = stim_target["img"][:, target_shifti::]
            stim_target["grating_mask"][:, 0 : cx - target_shifti] = stim_target["grating_mask"][
                :, target_shifti::
            ]
        elif distance_metric == "vertical":
            stim_target["img"][0 : cx - target_shifti, :] = stim_target["img"][target_shifti::, :]
            stim_target["grating_mask"][0 : cx - target_shifti, :] = stim_target["grating_mask"][
                target_shifti::, :
            ]
    else:
        if distance_metric == "horizontal":
            stim_target["img"][:, target_shifti::] = stim_target["img"][:, 0 : cx - target_shifti]
            stim_target["grating_mask"][:, target_shifti::] = stim_target["grating_mask"][
                :, 0 : cx - target_shifti
            ]
        elif distance_metric == "vertical":
            stim_target["img"][target_shifti::, :] = stim_target["img"][0 : cx - target_shifti, :]
            stim_target["grating_mask"][target_shifti::, :] = stim_target["grating_mask"][
                0 : cx - target_shifti, :
            ]

    # Add targets on grating
    mask_temp = np.ones(stim["shape"])
    mask_temp[stim_target["img"] == intensity_target] = 0
    img = stim["img"] * mask_temp + stim_target["img"]

    # Create target mask
    mask = np.where(stim_target["img"] == intensity_target, stim_target["grating_mask"], 0)
    unique_vals = np.unique(mask)
    for v in range(len(unique_vals)):
        mask[mask == unique_vals[v]] = v

    stim["img"] = img
    stim["target_mask"] = mask.astype(int)
    stim["target_phase_shift"] = target_phasei
    stim["target_size"] = stim_target["visual_size"]
    stim["intensity_target"] = intensity_target
    return stim


def grating_induction(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensities=(0.0, 1.0),
    target_width=None,
    intensity_target=0.5,
    origin="corner",
):
    """Grating induction illusion using a sinewave grating

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_bars : Sequence[float, ...]
        intensity value for each bar, by default (1.0, 0.0).
        Can specify as many intensities as n_bars;
        If fewer intensities are passed than n_bars, cycles through intensities
    target_width : float
        width of target stripe in degrees visual angle
    intensities : Sequence[float, float] or None (default)
        min and max intensity of sinewave
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    McCourt, M. E. (1982).
        A spatial frequency dependent grating-induction effect.
        Vision Research, 22, 119-134.
        https://doi.org/10.1016/0042-6989(82)90173-0
    """
    if target_width is None:
        raise ValueError("induction() missing argument 'target_width' which is not 'None'")

    # Draw grating
    stim = sinewave(
        shape=shape,
        visual_size=visual_size,
        ppd=ppd,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        intensities=intensities,
        origin=origin,
    )

    # Identify target region
    rectangle_size = (target_width, stim["visual_size"].width)

    target_mask = rectangle(
        rectangle_size=rectangle_size,
        ppd=stim["ppd"],
        visual_size=stim["visual_size"],
        intensity_background=0,
        intensity_rectangle=1,
    )

    # Superimpose
    stim["img"] = np.where(target_mask["rectangle_mask"], intensity_target, stim["img"])
    stim["target_mask"] = np.where(target_mask["rectangle_mask"], stim["grating_mask"], 0)
    return stim


def grating_induction_blur(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensity_bars=(0.0, 1.0),
    target_width=None,
    sigma=None,
    intensity_target=0.5,
    origin="corner",
):
    """Grating induction illusion using a blurred square-wave grating

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_bars : Sequence[float, ...]
        intensity value for each bar, by default (1.0, 0.0).
        Can specify as many intensities as n_bars;
        If fewer intensities are passed than n_bars, cycles through intensities
    target_width : float
        width of target stripe in degrees visual angle
    target_blur : float
        amount of Gaussian blur to blur square-wave grating (default: 0)
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
     McCourt, M. E. (1982).
        A spatial frequency dependent grating-induction effect.
        Vision Research, 22, 119-134.
        https://doi.org/10.1016/0042-6989(82)90173-0
    """
    if target_width is None:
        raise ValueError("induction_blur() missing argument 'target_width' which is not 'None'")

    # Draw grating
    stim = squarewave(
        shape=shape,
        visual_size=visual_size,
        ppd=ppd,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        intensity_bars=intensity_bars,
        origin=origin,
        round_phase_width=True,
    )

    gauss = gaussian(
        visual_size=stim["visual_size"],
        ppd=stim["ppd"],
        shape=stim["shape"],
        sigma=sigma,
        origin="center",
    )["img"]
    stim["img"] = convolve(stim["img"], gauss / np.sum(gauss), "same", padding=True)

    # Identify target region
    rectangle_size = (target_width, stim["visual_size"].width)

    target_mask = rectangle(
        rectangle_size=rectangle_size,
        ppd=stim["ppd"],
        visual_size=stim["visual_size"],
        intensity_background=0,
        intensity_rectangle=1,
    )

    # Superimpose
    stim["img"] = np.where(target_mask["rectangle_mask"], intensity_target, stim["img"])
    stim["target_mask"] = np.where(target_mask["rectangle_mask"], stim["grating_mask"], 0)
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    params = {
        "ppd": 40,
        "n_bars": 8,
        "bar_width": 1.0,
    }

    small_grating = {
        "ppd": 40,
        "bar_width": 1.0,
        "n_bars": 7,
        "intensity_bars": (0.2, 0.8),
        "target_indices": (0, 2, 4, 6),
        "rotation": 90,
    }

    large_grating = {
        "ppd": 40,
        "bar_width": 1.0,
        "n_bars": 21,
    }

    # fmt: off
    stimuli = {
        "grating_on_uniform": on_uniform(**params, visual_size=20, grating_size=5, target_indices=2),
        "grating_on_grating": on_grating(large_grating_params=large_grating, small_grating_params=small_grating),
        "grating_on_grating-masked": on_grating_masked(
            large_grating_params=large_grating,
            small_grating_params=small_grating,
            mask_size=(5, 5, 2),
        ),
        "grating_phase_shifted": phase_shifted(
            **params, target_size=4, target_phase_shift=90
        ),
        "grating_induction": grating_induction(**params, target_width=0.5),
        "grating_induction_blurred-squarewave": grating_induction_blur(**params, target_width=0.5, sigma=0.1),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import numpy as np

from stimupy.components import draw_regions, mask_regions

__all__ = [
    "frames",
]


def mask_frames(
    edges,
    shape=None,
    visual_size=None,
    ppd=None,
    origin="mean",
    rotation=0.0,
):
    """Generate mask with integer indices for sequential square frames

    Parameters
    ----------
    edges : Sequence[Number, ...]
        upper-limit of each frame, in degrees visual angle
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)

    Returns
    -------
    dict[str, Any]
        mask with integer index for each frame (key: "frame_mask"),
        and additional keys containing stimulus parameters
    """
    stim = mask_regions(
        distance_metric="rectilinear",
        edges=edges,
        rotation=rotation,
        shape=shape,
        visual_size=visual_size,
        ppd=ppd,
        origin=origin,
    )
    stim["frame_mask"] = stim.pop("mask")

    return stim


def frames(
    visual_size=None,
    ppd=None,
    shape=None,
    radii=None,
    rotation=0.0,
    intensity_frames=(0.0, 1.0),
    intensity_background=0.5,
    origin="mean",
):
    """Draw sequential set of square frames with specified radii

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radii : Sequence[Number]
        radii of each frame, in degrees visual angle
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_frames : Sequence[float, ...], optional
        intensity value for each frame, by default (1.0, 0.0).
        Can specify as many intensities as number of frame_widths;
        If fewer intensities are passed than frame_widhts, cycles through intensities
    intensity_background : float, optional
        intensity value of background, by default 0.5
    origin : "corner", "mean" or "center", optional
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each frame (key: "frame_mask"),
        and additional keys containing stimulus parameters
    """
    if radii is None:
        raise ValueError("frames() missing argument 'radii' which is not 'None'")

    if not isinstance(radii, (int, float)):
        if np.diff(radii).min() < 0:
            raise ValueError("radii need to monotonically increase")

    # Get frames mask
    stim = mask_frames(
        edges=radii,
        shape=shape,
        visual_size=visual_size,
        ppd=ppd,
        origin=origin,
        rotation=rotation,
    )

    # Draw image and update args
    stim["img"] = draw_regions(
        stim["frame_mask"], intensities=intensity_frames, intensity_background=intensity_background
    )
    stim["radii"] = radii
    stim["intensity_frames"] = intensity_frames
    stim["intensity_background"] = intensity_background
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": (10, 10),
        "ppd": 30,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "frames_frames": frames(**default_params, radii=(1, 2, 3)),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=True, save=None)



import collections
import copy
import functools
import itertools

import numpy as np
import scipy.special as sp

from stimupy.utils import resolution

__all__ = [
    "round_to_vals",
    "int_factorize",
    "get_function_argument_names",
    "apply_bessel",
    "resize_array",
    "resize_dict",
    "stack_dicts",
    "rotate_dict",
    "flip_dict",
    "roll_dict",
    "strip_dict",
    "make_two_sided",
    "permutate_params",
    "create_stimspace_stimuli",
]


def round_to_vals(arr, vals):
    """
    Round array to provided values (vals)

    Parameters
    ----------
    arr : np.ndarray
        Numpy array which values will be rounded
    vals : Sequence(float, ...)
        Values to which array will be rounded

    Returns
    -------
    out_arr : np.ndarray
        Rounded output array

    """
    n_val = len(vals)
    arr = np.repeat(np.expand_dims(arr, -1), n_val, axis=2)
    vals_arr = np.ones(arr.shape) * np.array(np.expand_dims(vals, [0, 1]))

    indices = np.argmin(np.abs(arr - vals_arr), axis=2)
    out_arr = np.copy(indices).astype(float)

    for i in range(n_val):
        out_arr[indices == i] = vals[i]
    return out_arr


def int_factorize(n):
    """All integer factors of integer n

    All integer factors, i.e., all integers that n is integer-divisible by.
    Also not a very efficient algorithm (brute force trial division),
    so should only be used as a helpter function.

    Parameters
    ----------
    n : int
        number to factorize

    Returns
    -------
    set
        set of all integer factors of n
    """

    factors = {1}  # set, guarantees unique factors
    for i in range(2, int(np.sqrt(n)) + 1):
        if n % i == 0:
            # N is divisible by i...
            factors.add(i)
            # ...thus also divisible by n/i
            factors.add(n // i)

    return factors


def get_function_argument_names(func):
    """
    Get all argument names for a given function

    Parameters
    ----------
    func : function
        Get argument names from this function

    Returns
    -------
    names : tuple
        Tuple containing all argument names of given function
    """
    names = func.__code__.co_varnames[: func.__code__.co_argcount]
    return names


def apply_bessel(arr, order=0):
    """
    Bessel function of the first kind of real order and complex argument.

    Parameters
    ----------
    arr : np.ndarray
        Input array
    order : float
        Order of the bessel function. Default is 0.

    Returns
    -------
    out : np.ndarray
        Output array

    """
    out = sp.jv(order, arr)
    return out


def resize_array(arr, factor):
    """
    Return a copy of an array, resized by the given factor. Every value is
    repeated factor[d] times along dimension d.

    Parameters
    ----------
    arr : 2D array
          the array to be resized
    factor : tuple of 2 ints
             the resize factor in the y and x dimensions

    Returns
    -------
    An array of shape (arr.shape[0] * factor[0], arr.shape[1] * factor[1])
    """
    return np.repeat(np.repeat(arr, factor[0], axis=0), factor[1], axis=1)


def resize_dict(dct, factor, keys=("img", "*mask")):
    """
    Return a copy of an array, resized by the given factor. Every value is
    repeated factor[d] times along dimension d.

    Parameters
    ----------
    dct : dict
        dict containing arrays to be resized
    factor : tuple of 2 ints
        the resize factor in the y and x dimensions
    keys : Sequence[String, String] or String
        keys in dict for images to be padded

    Returns
    -------
    dict[str, Any]
        same as input dict but with larger key-arrays according to
        "(arr.shape[0] * factor[0], arr.shape[1] * factor[1])"
        and updated keys for "visual_size" and "shape"
    """
    # Create deepcopy to not override existing dict
    new_dict = copy.deepcopy(dct)

    if isinstance(keys, str):
        keys = (keys,)

    # Find relevant keys
    keys = [
        dkey
        for key in keys
        for dkey in dct.keys()
        if ((dkey == key) or ((dkey.endswith(key[1::])) and (key.startswith("*"))))
    ]

    for key in dct.keys():
        if key in keys:
            img = dct[key]
            if isinstance(img, np.ndarray):
                img = np.repeat(np.repeat(img, factor[0], axis=0), factor[1], axis=1)
                if key.endswith("mask"):
                    img = img.astype(int)
                new_dict[key] = img

    # Update visual_size and shape-keys
    dct["shape"] = resolution.validate_shape(img.shape)
    if "ppd" in dct.keys():
        dct["visual_size"] = resolution.visual_size_from_shape_ppd(img.shape, dct["ppd"])
    return new_dict


def stack_dicts(
    dct1, dct2, direction="horizontal", keys=("img", "*mask"), keep_mask_indices=False
):
    """
    Return a dict with resized key-arrays by the given factor. Every value is
    repeated factor[d] times along dimension d.

    Parameters
    ----------
    dct1: dict
        dict containing arrays to be stacked
    dct2: dict
        dict containing arrays to be stacked
    direction : str
        stack horizontal(ly) or vertical(ly) (default: horizontal)
    keys : Sequence[String, String] or String
        keys in dict for images to be padded

    Returns
    -------
    dict[str, Any]
        same as input dict1 but with stacked key-arrays and updated keys for
        "visual_size" and "shape"
    """

    # Create deepcopy to not override existing dict
    new_dict = copy.deepcopy(dct1)

    if isinstance(keys, str):
        keys = (keys,)

    # Find relevant keys
    keys1 = [
        dkey
        for key in keys
        for dkey in dct1.keys()
        if ((dkey == key) or ((dkey.endswith(key[1::])) and (key.startswith("*"))))
    ]
    keys2 = [
        dkey
        for key in keys
        for dkey in dct2.keys()
        if ((dkey == key) or ((dkey.endswith(key[1::])) and (key.startswith("*"))))
    ]

    if not keys1 == keys2:
        raise ValueError("The requested keys do not exist in both dicts")

    for key in dct1.keys():
        if key in keys1:
            img1 = dct1[key]
            img2 = dct2[key]
            if isinstance(img1, np.ndarray) and isinstance(img2, np.ndarray):
                if key.endswith("mask") and not keep_mask_indices:
                    img2 = np.where(img2 != 0, img2 + img1.max(), 0)

                if direction == "horizontal":
                    img = np.hstack([img1, img2])
                elif direction == "vertical":
                    img = np.vstack([img1, img2])
                else:
                    raise ValueError("direction must be horizontal or vertical")

                if key.endswith("mask"):
                    img = img.astype(int)
                new_dict[key] = img

    # Update visual_size and shape-keys
    new_dict["shape"] = resolution.validate_shape(img.shape)
    if "ppd" in new_dict.keys():
        new_dict["visual_size"] = resolution.visual_size_from_shape_ppd(img.shape, new_dict["ppd"])
    return new_dict


def rotate_dict(dct, nrots=1, keys=("img", "*mask")):
    """
    Return a dict with key-arrays rotated by nrots*90 degrees.

    Parameters
    ----------
    dct: dict
        dict containing arrays to be stacked
    nrot : int
        number of rotations by 90 degrees
    keys : Sequence[String, String] or String
        keys in dict for images to be padded

    Returns
    -------
    dict[str, Any]
        same as input dict but with rotated key-arrays and updated keys for
        "visual_size" and "shape"
    """

    # Create deepcopy to not override existing dict
    new_dict = copy.deepcopy(dct)

    # Find relevant keys
    keys = [
        dkey
        for key in keys
        for dkey in dct.keys()
        if ((dkey == key) or ((dkey.endswith(key[1::])) and (key.startswith("*"))))
    ]

    for key in dct.keys():
        if key in keys:
            img = dct[key]
            if isinstance(img, np.ndarray):
                if isinstance(nrots, (int, float)):
                    img = np.rot90(img, nrots)
                else:
                    raise ValueError("nrots must be a number")

                if key.endswith("mask"):
                    img = img.astype(int)
                new_dict[key] = img

    # Update visual_size and shape-keys
    new_dict["shape"] = resolution.validate_shape(img.shape)
    if "ppd" in new_dict.keys():
        new_dict["visual_size"] = resolution.visual_size_from_shape_ppd(img.shape, new_dict["ppd"])
    return new_dict


def flip_dict(dct, direction="lr", keys=("img", "*mask")):
    """
    Return a dict with key-arrays rotated by nrots*90 degrees.

    Parameters
    ----------
    dct: dict
        dict containing arrays to be stacked
    direction : str
        "lr" for left-right, "ud" for up-down flipping
    keys : Sequence[String, String] or String
        keys in dict for images to be padded

    Returns
    -------
    dict[str, Any]
        same as input dict but with flipped key-arrays
    """

    # Create deepcopy to not override existing dict
    new_dict = copy.deepcopy(dct)

    # Find relevant keys
    keys = [
        dkey
        for key in keys
        for dkey in dct.keys()
        if ((dkey == key) or ((dkey.endswith(key[1::])) and (key.startswith("*"))))
    ]

    for key in dct.keys():
        if key in keys:
            img = dct[key]
            if isinstance(img, np.ndarray):
                if direction == "lr":
                    img = np.fliplr(img)
                elif direction == "ud":
                    img = np.flipud(img)
                else:
                    raise ValueError("direction must be lr or ud")

                if key.endswith("mask"):
                    img = img.astype(int)
                new_dict[key] = img
    return new_dict


def roll_dict(dct, shift, axes, keys=("img", "*mask")):
    """
    Return a dict with key-arrays rolled by shift in axes.

    Parameters
    ----------
    dct: dict
        dict containing arrays to be stacked
    shift : int
        number of pixels by which to shift
    axes : Number or Sequence[Number, ...]
        axes in which to shift
    keys : Sequence[String, String] or String
        keys in dict for images to be padded

    Returns
    -------
    dict[str, Any]
        same as input dict but with rolled key-arrays
    """

    # Create deepcopy to not override existing dict
    new_dict = copy.deepcopy(dct)
    shift = np.array(shift).astype(int)

    # Find relevant keys
    keys = [
        dkey
        for key in keys
        for dkey in dct.keys()
        if ((dkey == key) or ((dkey.endswith(key[1::])) and (key.startswith("*"))))
    ]

    for key in dct.keys():
        if key in keys:
            img = dct[key]
            if isinstance(img, np.ndarray):
                img = np.roll(img, shift=shift, axis=axes)

                if key.endswith("mask"):
                    img = img.astype(int)
                new_dict[key] = img
    return new_dict


def strip_dict(
    dct,
    func,
):
    """
    Create a dictionary by stripping it from all keys that are not also
    an argument to the provided function

    Parameters
    ----------
    dct: dict
        dict which will be stripped
    func : function
        Get argument names from this function

    Returns
    -------
    dict[str, Any]
        same as input dict but stripped from all keys which are not also
        an argument to the provided function
    """

    # Get all argument names for given function
    names = get_function_argument_names(func)

    # Add keys from dct into new_dict if the key is also an argument name
    new_dict = dict()
    for name in names:
        if name in dct.keys():
            new_dict[name] = dct[name]
    return new_dict


def make_two_sided(func, two_sided_params):
    """Create two-sided version of a stimulus function

    Where (some) parameters can be specified separately for each side.
    These parameters should then be specified as a 2-Sequence (2-ple, list of len=2),
    where entry [0] is the parameter value for left side, and [1] for right side.
    This means that if the kwarg takes a Sequence itself, e.g., `intensities`,
    then the two-sided specification must be, e.g.,
    ((int_left_0, int_left_1), (int_right_0, int_right_1))

    Will be left- and right-sided.


    Parameters
    ----------
    func : function
        stimulus function to double
    two_sided_params : Sequence[str]
        names of parameters (kwargs) of func
        that can be specified separately for each side of the display.

    Returns
    -------
    function
        two-sided version of stimulus function
    """

    @functools.wraps(func)
    def two_sided_func(**kwargs):
        shape = kwargs.pop("shape", None)
        visual_size = kwargs.pop("visual_size", None)
        ppd = kwargs.pop("ppd", None)

        # Resolve resolution
        try:
            shape, visual_size, ppd = resolution.resolve(
                shape=shape, visual_size=visual_size, ppd=ppd
            )
        except resolution.TooManyUnknownsError:
            shape = resolution.validate_shape(shape)
            visual_size = resolution.validate_visual_size(visual_size)
            ppd = resolution.validate_ppd(ppd)
        if visual_size.width is not None:
            _visual_size = (visual_size.height, visual_size.width / 2)
        else:
            _visual_size = None
        if shape.width is not None:
            _shape = (shape.height, shape.width / 2)
        else:
            _shape = None

        # Left side
        left_side_params = {}
        for key, value in kwargs.items():
            if key in two_sided_params and isinstance(value, collections.abc.Sequence):
                left_side_params[key] = value[0]
            else:
                left_side_params[key] = value

        left = func(
            visual_size=_visual_size,
            ppd=ppd,
            shape=_shape,
            **left_side_params,
        )

        # Right side
        right_side_params = {}
        for key, value in kwargs.items():
            if key in two_sided_params and isinstance(value, collections.abc.Sequence):
                right_side_params[key] = value[1]
            else:
                right_side_params[key] = value

        right = func(
            visual_size=_visual_size,
            ppd=ppd,
            shape=_shape,
            **right_side_params,
        )

        stim = stack_dicts(left, right, direction="horizontal")
        stim["shape"] = left["shape"].height, left["shape"].width + right["shape"].width
        stim["visual_size"] = (
            left["visual_size"].height,
            left["visual_size"].width + right["visual_size"].width,
        )
        for key in two_sided_params:
            if key in kwargs:
                stim[key] = kwargs[key]
            else:
                stim[key] = (left[key], right[key])
        return stim

    return two_sided_func


def permutate_params(params):
    if not isinstance(params, dict):
        raise ValueError("params needs to be a dict with all stimulus parameters")

    keys, values = zip(*params.items())
    permutations_dicts = [dict(zip(keys, v)) for v in itertools.product(*values)]
    return permutations_dicts


def create_stimspace_stimuli(stimulus_function, permutations_dicts, title_params=None):
    if not callable(stimulus_function):
        raise ValueError("stimulus_function needs to be a function")
    if isinstance(title_params, str):
        title_params = [
            title_params,
        ]

    stimuli = {}
    for i, p in enumerate(permutations_dicts):
        if title_params is None:
            key = str(i)
        else:
            key = ""
            for tname in title_params:
                if isinstance(p[tname], (float, int)):
                    ptname = np.round(p[tname], 3)
                    key += f"{tname}={ptname} "
                else:
                    key += f"{tname}={p[tname]} "
        stimuli[key] = stimulus_function(
            **p,
        )
    return stimuli
